"""Deserializer for an enum."""

import enum
import logging
from dataclasses import dataclass
from typing import Any, Generic, Iterable, List, Sequence, Tuple, TypeVar, Union

from typing_extensions import Final, Literal, Protocol, final

from . import pretty_printer, type_utils
from .core import ConstructorCore, FactoryCore, InternalFactory
from .data import Value
from .errors import ConstructorError, UnsupportedTargetTypeError
from .types import Target

# We should really support bool type literals, too, but it is surprisingly
# tricky to tell the difference between Literal[1] and Literal[True] (and
# similarly Literal[0] and Literal[False]).
# The behaviour varies between Python versions and even depending on execution
# order. In a Python 3.8 interpreter session, for example, the value of
# Literal[1].__args__ and Literal[True].__args__ is either (True,) or (1,)
# depending on which was evaluated first!
_SUPPORTED_TYPES: Final[Sequence[type]] = [type(None), int, str]

_log = logging.getLogger(__name__)


_VariantType = TypeVar("_VariantType", None, int, str)
_T_co = TypeVar("_T_co", covariant=True)


@final
@dataclass(frozen=True)
class EnumFactory(FactoryCore):
    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if not (isinstance(target, type) and issubclass(target, enum.Enum)):
            raise UnsupportedTargetTypeError
        variants: Sequence[_EnumVariant] = list(target)
        _log.debug("Extracted enum variants %s", variants)
        for variant in variants:
            if type(variant.value) not in _SUPPORTED_TYPES:
                raise factory.add_context(
                    type(variant.value), context=f"variant {variant.name}"
                ).make_error(
                    "only enum values of type None, bool, int, and str are supported"
                )
        return _Constructor(
            _variant_constructors=[
                _VariantConstructorImpl(
                    v,
                    v.value,
                    factory.create_constructor(
                        type(v.value), context=f"variant {v.name}"
                    ).unwrap(),
                )
                for v in variants
            ],
        )


@final
@dataclass(frozen=True)
class LiteralFactory(FactoryCore):
    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if type_utils.get_base_of_generic_type(target) is not Literal:
            raise UnsupportedTargetTypeError
        variants = type_utils.get_type_parameters(target)
        _log.debug("Extracted literal variants %s", variants)
        for variant in variants:
            if not any(isinstance(variant, t) for t in _SUPPORTED_TYPES):
                raise factory.add_context(
                    type(variant),
                    context=f"variant {variant}",
                ).make_error(
                    "only literal variants of type None, int, and str are supported"
                )
        return _Constructor(
            _variant_constructors=[
                _VariantConstructorImpl(
                    variant,
                    variant,
                    factory.create_constructor(
                        type(variant), context=f"variant {variant}"
                    ).unwrap(),
                )
                for variant in variants
            ],
        )


@dataclass(frozen=True)
class _VariantConstructorImpl(Generic[_VariantType]):
    variant: _VariantType
    value: Any
    constructor: ConstructorCore[_VariantType]


@dataclass(frozen=True)
class _Constructor(ConstructorCore[_T_co]):
    _variant_constructors: Sequence[_VariantConstructorImpl[Any]]

    def __call__(self, data: Value) -> _T_co:
        possible_variants: List[Tuple[_VariantConstructorImpl[Any], Any]] = []
        for variant in self._variant_constructors:
            try:
                possible_variants.append((variant, variant.constructor(data)))
            except ConstructorError:
                pass

        if not possible_variants:
            coercing_strings_str = ""
            if data.config.coerce_strings:
                coercing_strings_str = "(string representation of) "
            variant_types_str = _prepend_one_of(
                {
                    pretty_printer.print_type_name(type(variant.value))
                    for variant in self._variant_constructors
                }
            )
            raise data.make_error(f"expected {coercing_strings_str}{variant_types_str}")

        for variant, value in possible_variants:
            if variant.value == value:
                return variant.variant

        coercing_strings_str = ""
        if data.config.coerce_strings:
            coercing_strings_str = "(string representation of) "
        variants_str = _prepend_one_of(
            _quote_literal_variant(v.value) for v in self._variant_constructors
        )
        raise data.make_error(
            f"unknown variant - expected {coercing_strings_str}{variants_str}"
        )


def _prepend_one_of(elements: Iterable[str]) -> str:
    elements = list(elements)
    if len(elements) == 1:
        return elements[0]
    return f"one of {{{','.join(sorted(elements))}}}"


def _quote_literal_variant(variant: Union[None, bool, int, str]) -> str:
    if variant is None or isinstance(variant, (bool, int, bytes)):
        return str(variant)
    return variant.replace("\\", "\\\\").replace('"', '\\"')


class _EnumVariant(Protocol):
    name: str
    value: Any



"""Deserializer for a union of types."""

import logging
from dataclasses import dataclass
from typing import List, Sequence, TypeVar, Union

from typing_extensions import final

from . import type_utils
from .core import ConstructorCore, FactoryCore, InternalConstructor, InternalFactory
from .data import Value
from .errors import ConstructorError, UnsupportedTargetTypeError
from .types import Target

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)


@final
@dataclass(frozen=True)
class UnionFactory(FactoryCore):
    @dataclass(frozen=True)
    class _Constructor(ConstructorCore[_T_co]):
        _variant_constructors: Sequence[InternalConstructor[_T_co]]

        def __call__(self, data: "Value") -> _T_co:
            errors: List[ConstructorError] = []
            for index, constructor in enumerate(self._variant_constructors):
                try:
                    _log.debug("Trying union variant %d: %s", index, constructor)
                    value = constructor(data)
                    _log.debug("Successfully constructed union variant %d", index)
                    return value
                except ConstructorError as e:
                    _log.debug("Constructor error for union variant %d: %s", index, e)
                    errors.append(e)
            raise sorted(errors, reverse=True)[0]

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if type_utils.get_base_of_generic_type(target) is not Union:
            raise UnsupportedTargetTypeError
        return type(self)._Constructor(
            [
                factory.create_constructor(
                    variant_type, context=f"union variant {index}"
                )
                for index, variant_type in enumerate(
                    type_utils.get_type_parameters(target)
                )
            ]
        )



import abc
import dataclasses
import logging
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    overload,
)

from typing_extensions import final

from .data import ConstructionConfig, Context, Value
from .errors import FactoryError, InternalError, TerramareError
from .pretty_printer import (
    LoggablePrimitive,
    LoggableTarget,
    print_table,
    print_type_name,
)
from .safe_mapping import SafeMutableMapping
from .types import Primitive, Target

if TYPE_CHECKING:  # pragma: no cover
    from .metadata import TargetMetadata

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)


@dataclass(frozen=True)
class InternalFactoryError(TerramareError):
    """Internal exception raised when failing to create a constructor."""

    summary: str
    detail: str


class ConstructorCore(Generic[_T_co], abc.ABC):
    """
    Interface for constructor implementations.

    This should not be used directly - instead, use:
    - the InternalConstructor wrapper class, if constructing a sub-object from
      another constructor;
    - the public Constructor class, if constructing a top-level object.
    """

    @abc.abstractmethod
    def __call__(self, data: "Value") -> _T_co:
        """
        Create an instance of the target type.

        :param data: Primitive value from which to construct an instance of the
            target type.

        :returns: Instance of the target type.

        :raises terramare.errors.ConstructorError: If the primitive value has
            an incorrect type or structure.
        :raises terramare.errors.ValidationException: If the primitive value has
            the correct type and structure but fails value-level validation.
        """


@final
@dataclass
class ForwardRef(ConstructorCore[_T_co]):
    """
    Forward reference to a ConstructorCore.

    This is necessary to create constructors for recursive types. For example,
    consider:

    @dataclass
    class LinkedList:
        next: Optional["LinkedList"] = None

    With a naive implementation we would not be able to create a LinkedList
    constructor without already having a LinkedList constructor!

    A ForwardRef can be returned in place of a constructor from
    `create_constructor` when it is called with a target type from within a
    call to `create_constructor` for that type.

    All instances of the ForwardRef are then resolved to a real constructor
    before returning from the earlier call to `create_constructor`.

    For example, calling `create_constructor(LinkedList)` as part of creating a
    constructor for the `next` field of LinkedList will return a ForwardRef,
    preventing infinite recursion.

    By the time a constructor is returned from the top-level `create_constructor`
    call this ForwardRef has been resolved - the same constructor is used for
    constructing a top-level LinkedList instance _and_ its `next` member.
    """

    _impl: Optional[ConstructorCore[_T_co]] = None

    def __call__(self, data: "Value") -> _T_co:
        """
        Call the referenced constructor.

        See `ConstructorCore.__call__`.

        :raises InternalError: If the ForwardRef has not been resolved.
        """
        return self.deref()(data)

    def resolve_to(self, impl: ConstructorCore[_T_co]) -> None:
        """Resolve the forward reference to the given real constructor."""
        self._impl = impl

    def deref(self) -> ConstructorCore[_T_co]:
        """
        Retrieve the referenced constructor.

        :raises InternalError: If the ForwardRef has not been resolved.
        """
        if self._impl is None:  # pragma: no cover
            raise InternalError(
                f"attempted to dereference unresolved forward reference {id(self):x}"
            )
        return self._impl


@final
@dataclass
class InternalConstructor(Generic[_T_co]):
    """
    Internal-only wrapper class for constructor implementations.

    Provides logging, error context, and forward reference handling.
    """

    _impl: ConstructorCore[_T_co]
    _target: Target[_T_co]

    def __call__(self, data: "Value") -> _T_co:
        """
        Create an instance of the target type.

        See `ConstructorCore.__call__`.
        """
        if isinstance(self._impl, ForwardRef):
            self._impl = self._impl.deref()
        _log.debug(
            "Attempting to construct %s from data: %s",
            LoggableTarget(self._target),
            LoggablePrimitive(data.raw),
        )
        value = self._impl(data.clone().push_type(print_type_name(self._target)))
        _log.debug(
            "Constructed %s: %s",
            LoggableTarget(self._target),
            value,
        )
        return value

    def __str__(self) -> str:  # pragma: no cover
        """Return a human-readable representation of the constructor."""
        return f"Constructor[{print_type_name(self._target)}]"

    def unwrap(self) -> ConstructorCore[_T_co]:
        """Retrieve the wrapped ConstructorCore."""
        return self._impl


class FactoryCore(abc.ABC):
    """
    Interface for constructor factory implementations.

    This should not be used directly - instead, use:
    - the InternalFactory wrapper class, if creating a sub-constructor from
      another constructor factory;
    - the public Factory class, if creating a top-level constructor.
    """

    @abc.abstractmethod
    def create_constructor(
        self, factory: "InternalFactory", target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        """
        Create a constructor for the given target type.

        :param factory: Constructor factory used when recursively creating sub-
            constructors.
        :param target: Target type.

        :returns: A `ConstructorCore` targeting the given type.

        :raises InternalFactoryError: If a constructor for the given target type
            cannot be created due to fundamental problems with the target type.
        :raises UnsupportedTargetTypeError: If a constructor for the given
            target type cannot be created by this factory, but another factory
            may be able to create such a constructor.
        """


@final
@dataclass(frozen=True)
class InternalFactory:
    """
    Internal-only wrapper class for constructor factory implementations.

    Provides logging, error context, caching and forward reference handling.

    A new instance of this class is created with each `create_constructor` call,
    including recursive calls. New instances share the persistent data of the
    parent instance.
    """

    _persistent_data: "PersistentData"
    _context: Sequence[Tuple[str, str]] = field(default_factory=list)

    @dataclass(frozen=True)
    class PersistentData:
        """Persistent data shared between recursive `create_constructor` calls."""

        impl: FactoryCore
        cache: SafeMutableMapping[Target[Any], InternalConstructor[Any]] = field(
            default_factory=SafeMutableMapping
        )

    def create_constructor(
        self, target: Target[_T_co], *, context: str = ""
    ) -> InternalConstructor[_T_co]:
        """
        Create a constructor for the given target type.

        :param target: Target type.
        :param context: Additional context for the constructor, such as an enum
            variant index.

        :returns: An `InternalConstructor` targeting the given type.

        :raises InternalFactoryError: If a constructor for the given target type
            cannot be created due to fundamental problems with the target type.
        :raises UnsupportedTargetTypeError: If a constructor for the given
            target type cannot be created by this factory, but another factory
            may be able to create such a constructor.
        """
        _log.debug(
            "Attempting to create constructor for target %s",
            LoggableTarget(target),
        )

        def create() -> ConstructorCore[_T_co]:
            return self._persistent_data.impl.create_constructor(
                self.add_context(target, context=context), target
            )

        if target not in self._persistent_data.cache:
            _log.debug("Cache miss for target %s", LoggableTarget(target))
            forward_ref: ForwardRef[_T_co] = ForwardRef()
            _log.debug(
                "Created ForwardRef %x for target %s",
                id(forward_ref),
                LoggableTarget(target),
            )
            self._persistent_data.cache[target] = InternalConstructor(
                forward_ref, target
            )
            try:
                constructor = create()
            except BaseException:
                # Remove the unresolved forward reference.
                del self._persistent_data.cache[target]
                raise
            forward_ref.resolve_to(constructor)
            _log.debug(
                "ForwardRef %x for target %s resolved to %s",
                id(forward_ref),
                LoggableTarget(target),
                constructor,
            )
            _log.debug("Cached constructor for target %s", LoggableTarget(target))
        _log.debug(
            "Retrieved constructor from cache for target %s",
            LoggableTarget(target),
        )
        return self._persistent_data.cache[target]

    def make_error(self, msg: str) -> InternalFactoryError:
        """
        Create an `InternalFactoryError` with the given message.

        The error will contain additional context including the stack of target
        types.
        """
        return InternalFactoryError(
            msg, print_table([("path", "type"), *self._context])
        )

    def add_context(
        self, target: Target[_T_co], *, context: str = ""
    ) -> "InternalFactory":
        return dataclasses.replace(
            self, _context=[*self._context, (context, print_type_name(target))]
        )


@final
@dataclass(frozen=True)
class Factory:
    _persistent_data: InternalFactory.PersistentData = field(
        # pylint: disable=unnecessary-lambda
        default_factory=lambda: Factory._make_persistent_data()
    )

    @classmethod
    def new(
        cls,
        *,
        _experimental_metadata: Optional[Mapping[Target[Any], "TargetMetadata"]] = None,
    ) -> "Factory":
        return cls(
            cls._make_persistent_data(_experimental_metadata=_experimental_metadata)
        )

    @overload
    def create_constructor(self, target: Target[_T_co]) -> "Constructor[_T_co]":
        # See terramare.py for an explanation of the overloads here.
        ...  # pragma: no cover

    @overload
    def create_constructor(self, target: Any) -> "Constructor[Any]":
        ...  # pragma: no cover

    def create_constructor(self, target: Any) -> "Constructor[_T_co]":
        factory = InternalFactory(self._persistent_data)
        try:
            constructor = factory.create_constructor(target)
        except InternalFactoryError as e:
            raise FactoryError(
                f"failed to create constructor for type '{print_type_name(target)}': "
                f"{e.summary}\n{e.detail}"
            ) from e
        return Constructor(constructor)

    @overload
    def structure(
        self,
        data: Primitive,
        *,
        into: Target[_T_co],
        coerce_strings: bool = False,
        context: Optional[Mapping[str, object]] = None,
    ) -> _T_co:
        # See terramare.py for an explanation of the overloads here.
        ...  # pragma: no cover

    @overload
    def structure(
        self,
        data: Primitive,
        *,
        into: Any,
        coerce_strings: bool = False,
        context: Optional[Mapping[str, object]] = None,
    ) -> Any:
        ...  # pragma: no cover

    def structure(
        self,
        data: Primitive,
        *,
        into: Any,
        coerce_strings: bool = False,
        context: Optional[Mapping[str, object]] = None,
    ) -> Any:
        return self.create_constructor(into)(
            data, coerce_strings=coerce_strings, context=context
        )

    @staticmethod
    def _make_persistent_data(
        *,
        _experimental_metadata: Optional[Mapping[Target[Any], "TargetMetadata"]] = None,
    ) -> InternalFactory.PersistentData:
        # pylint: disable=import-outside-toplevel, cyclic-import
        from .classes import ClassFactory
        from .composite import CompositeFactory
        from .context_factory import ContextFactory
        from .enums import EnumFactory, LiteralFactory
        from .mappings import MappingFactory, TypedDictFactory
        from .metadata import MetadataCollection, MetadataFactory
        from .newtypes import NewTypeFactory
        from .primitives import (
            AnyFactory,
            BoolFactory,
            FloatFactory,
            IntFactory,
            NoneFactory,
            PrimitiveDictFactory,
            PrimitiveListFactory,
            StrFactory,
        )
        from .sequences import SequenceFactory, TupleFactory
        from .stdlib_metadata import DEFAULT_METADATA
        from .tagged_constructors import TaggedFactory
        from .unions import UnionFactory

        metadata = MetadataCollection.new(
            {**DEFAULT_METADATA, **(_experimental_metadata or {})}
        )

        return InternalFactory.PersistentData(
            CompositeFactory(
                [
                    ContextFactory(),
                    MetadataFactory(metadata),
                    TaggedFactory(metadata),
                    NewTypeFactory(),
                    NoneFactory(),
                    BoolFactory(),
                    IntFactory(),
                    FloatFactory(),
                    StrFactory(),
                    PrimitiveListFactory(),
                    PrimitiveDictFactory(),
                    AnyFactory(),
                    EnumFactory(),
                    LiteralFactory(),
                    SequenceFactory(),
                    TupleFactory(),
                    MappingFactory(),
                    TypedDictFactory(),
                    UnionFactory(),
                    ClassFactory(metadata),
                ]
            )
        )


@final
@dataclass(frozen=True)
class Constructor(Generic[_T_co]):
    _impl: InternalConstructor[_T_co]

    def __call__(
        self,
        data: Primitive,
        *,
        coerce_strings: bool = False,
        context: Optional[Mapping[str, object]] = None,
    ) -> _T_co:
        if not _is_primitive(data):
            raise TypeError("unsupported type for 'data': expected Primitive")
        if not isinstance(coerce_strings, bool):
            raise TypeError("unsupported type for 'coerce_strings': expected bool")
        return self._impl(
            Value.new(
                data,
                config=ConstructionConfig(coerce_strings=coerce_strings),
                context=Context(context or {}),
            )
        )


def _is_primitive(data: Any) -> bool:
    if data is None:
        return True
    if isinstance(data, (bool, int, float, str)):
        return True
    if isinstance(data, list):
        return all(_is_primitive(element) for element in data)
    if isinstance(data, dict):
        return all(
            isinstance(key, str) and _is_primitive(value) for key, value in data.items()
        )
    return False



from dataclasses import dataclass
from typing import Any, ClassVar, Generic, Mapping, Optional, Type, TypeVar

from typing_extensions import final

from .core import ConstructorCore, FactoryCore, InternalConstructor, InternalFactory
from .data import Value
from .errors import UnsupportedTargetTypeError
from .metadata import Metadata, MetadataCollection
from .types import Lazy, Target

_T_co = TypeVar("_T_co", covariant=True)


@dataclass(frozen=True)
class externally_tagged(Metadata[Optional[Lazy[Mapping[str, Target[Any]]]]]):
    """
    Mark the decorated class as an externally-tagged variant type.

    >>> import terramare
    >>> from dataclasses import dataclass
    >>>
    >>> @terramare.externally_tagged(lambda: {"request": Request, "response": Response})
    ... class Message:
    ...     pass
    >>>
    >>> @dataclass
    ... class Request:
    ...     method: str
    >>>
    >>> @dataclass
    ... class Response:
    ...     code: int
    >>>
    >>> terramare.structure({"request": {"method": "GET"}}, into=Message)
    Request(method='GET')
    """

    KEY: ClassVar[str] = f"{__name__}.externally_tagged"
    DEFAULT: ClassVar[Optional[Lazy[Mapping[str, Target[Any]]]]] = None

    _value: Lazy[Mapping[str, Target[Any]]]

    @property
    def value(self) -> Lazy[Mapping[str, Target[Any]]]:
        return self._value


@dataclass(frozen=True)
class internally_tagged(Metadata[Optional["internally_tagged"]]):
    """
    Mark the decorated class as an internally-tagged variant type.

    >>> import terramare
    >>> from dataclasses import dataclass
    >>>
    >>> @terramare.internally_tagged(key="type", variants=lambda: {
    ...     "request": Request, "response": Response
    ... })
    ... class Message:
    ...     pass
    >>>
    >>> @dataclass
    ... class Request:
    ...     method: str
    >>>
    >>> @dataclass
    ... class Response:
    ...     code: int
    >>>
    >>> terramare.structure({"type": "request", "method": "GET"}, into=Message)
    Request(method='GET')
    """

    KEY: ClassVar[str] = f"{__name__}.internally_tagged"
    DEFAULT: ClassVar[Optional["internally_tagged"]] = None

    key: str
    variants: Lazy[Mapping[str, type]]
    default_variant: Optional[str] = None

    def __init__(
        self,
        key: str,
        variants: Lazy[Mapping[str, type]],
        *,
        default_variant: Optional[str] = None,
    ) -> None:
        # Define a custom __init__ function so that 'default_variant' can be
        # made keyword-only.
        object.__setattr__(self, "key", key)
        object.__setattr__(self, "variants", variants)
        object.__setattr__(self, "default_variant", default_variant)

    @property
    def value(self) -> "internally_tagged":
        return self


@final
@dataclass(frozen=True)
class TaggedFactory(FactoryCore):
    _metadata: MetadataCollection

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        externally_tagged_variants = externally_tagged.read(self._metadata, target)
        if externally_tagged_variants:
            return ExternallyTaggedConstructor.new(
                factory, externally_tagged_variants()
            )

        internally_tagged_variants = internally_tagged.read(self._metadata, target)
        if internally_tagged_variants:
            return InternallyTaggedConstructor.new(
                factory,
                internally_tagged_variants.key,
                internally_tagged_variants.variants(),
                default_variant=internally_tagged_variants.default_variant,
            )

        raise UnsupportedTargetTypeError


@final
@dataclass(frozen=True)
class ExternallyTaggedConstructor(ConstructorCore[_T_co]):
    _variants: Mapping[str, InternalConstructor[_T_co]]

    @classmethod
    def new(
        cls,
        factory: InternalFactory,
        variants: Mapping[str, Target[_T_co]],
    ) -> "ExternallyTaggedConstructor[_T_co]":
        return cls(
            {
                tag: factory.create_constructor(
                    variant, context=f'externally-tagged variant "{tag}"'
                )
                for tag, variant in variants.items()
            },
        )

    def __call__(self, data: Value) -> _T_co:
        object_data = data.as_object()
        if not object_data:
            raise data.make_error("missing variant")
        if len(object_data) > 1:
            raise data.make_error("multiple variants")
        tag = next(iter(object_data.keys()))
        if tag in self._variants:
            return self._variants[tag](object_data[tag])
        variants_str = ",".join(sorted(self._variants))
        raise data.make_error(
            f"unknown variant ({tag}) - expected one of {{{variants_str}}}"
        )


@final
@dataclass(frozen=True)
class InternallyTaggedConstructor(Generic[_T_co], ConstructorCore[_T_co]):
    _tag_field: str
    _tag_constructor: InternalConstructor[str]
    _variants: Mapping[str, InternalConstructor[_T_co]]
    _default_variant: Optional[str] = None

    @classmethod
    def new(
        cls,
        factory: InternalFactory,
        tag_field: str,
        variants: Mapping[str, Type[_T_co]],
        *,
        default_variant: Optional[str] = None,
    ) -> "InternallyTaggedConstructor[_T_co]":
        if not (default_variant is None or default_variant in variants):
            variants_str = ",".join(sorted(variants))
            raise factory.make_error(
                f"unknown default variant ({default_variant}) - "
                f"expected one of {{{variants_str}}}"
            )
        return cls(
            tag_field,
            factory.create_constructor(str),
            {
                tag: factory.create_constructor(
                    variant, context=f'internally-tagged variant "{tag}"'
                )
                for tag, variant in variants.items()
            },
            _default_variant=default_variant,
        )

    def __call__(self, data: Value) -> _T_co:
        object_data = data.as_object()
        if self._tag_field in object_data:
            tag = self._tag_constructor(object_data[self._tag_field])
            del object_data[self._tag_field]
        elif self._default_variant is not None:
            tag = self._default_variant
        else:
            raise data.make_error(f"missing required field: {self._tag_field}")

        if tag in self._variants:
            return self._variants[tag](object_data)
        variants_str = ",".join(sorted(str(v) for v in self._variants))
        raise data.make_error(
            f"unknown variant ({tag}) - expected one of {{{variants_str}}}"
        )



"""Deserializers for primitive types."""

from dataclasses import dataclass
from typing import Any, Dict, List, Mapping, TypeVar, cast

from typing_extensions import Final, final

from .core import ConstructorCore, FactoryCore, InternalFactory
from .data import Value
from .errors import UnsupportedTargetTypeError
from .types import Primitive, Target

_T_co = TypeVar("_T_co", covariant=True)


_BOOL_STRINGS: Final[Mapping[str, bool]] = {
    "y": True,
    "Y": True,
    "yes": True,
    "Yes": True,
    "YES": True,
    "n": False,
    "N": False,
    "no": False,
    "No": False,
    "NO": False,
    "true": True,
    "True": True,
    "TRUE": True,
    "false": False,
    "False": False,
    "FALSE": False,
    "on": True,
    "On": True,
    "ON": True,
    "off": False,
    "Off": False,
    "OFF": False,
}


@final
@dataclass(frozen=True)
class NoneFactory(FactoryCore):
    class _Constructor(ConstructorCore[None]):
        def __call__(self, data: Value) -> None:
            if data.raw is not None:
                raise data.make_error("expected null")
            return data.raw

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not type(None):  # noqa: E721
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())


@final
@dataclass(frozen=True)
class BoolFactory(FactoryCore):
    @dataclass(frozen=True)
    class _Constructor(ConstructorCore[bool]):
        values: Mapping[str, bool]

        def __call__(self, data: Value) -> bool:
            if data.config.coerce_strings:
                return self._construct_coerce(data)
            else:
                return self._construct_no_coerce(data)

        def _construct_no_coerce(self, data: Value) -> bool:
            if isinstance(data.raw, bool):
                return data.raw
            raise data.make_error("expected bool")

        def _construct_coerce(self, data: Value) -> bool:
            if isinstance(data.raw, str):
                if data.raw in self.values:
                    return self.values[data.raw]
            values_str = "|".join(self.values)
            raise data.make_error(
                f"expected (string representation of) bool ({values_str})"
            )

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not bool:
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor(_BOOL_STRINGS))


@final
@dataclass(frozen=True)
class IntFactory(FactoryCore):
    class _Constructor(ConstructorCore[int]):
        def __call__(self, data: Value) -> int:
            if data.config.coerce_strings:
                return self._construct_coerce(data)
            else:
                return self._construct_no_coerce(data)

        def _construct_no_coerce(self, data: Value) -> int:
            # bool is a subclass of int - we want to make sure what we have is
            # actually an int.
            if isinstance(data.raw, int) and not isinstance(data.raw, bool):
                return data.raw
            raise data.make_error("expected integer")

        def _construct_coerce(self, data: Value) -> int:
            if isinstance(data.raw, str):
                try:
                    return int(data.raw)
                except ValueError:
                    pass
            raise data.make_error("expected (string representation of) integer")

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not int:
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())


@final
@dataclass(frozen=True)
class FloatFactory(FactoryCore):
    class _Constructor(ConstructorCore[float]):
        def __call__(self, data: Value) -> float:
            if data.config.coerce_strings:
                return self._construct_coerce(data)
            else:
                return self._construct_no_coerce(data)

        def _construct_no_coerce(self, data: Value) -> float:
            if isinstance(data.raw, float):
                return data.raw

            # Both floats and ints are acceptable.
            if isinstance(data.raw, int):
                return float(data.raw)
            raise data.make_error("expected number")

        def _construct_coerce(self, data: Value) -> float:
            if isinstance(data.raw, str):
                try:
                    return float(data.raw)
                except ValueError:
                    pass
            raise data.make_error("expected (string representation of) number")

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not float:
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())


@final
@dataclass(frozen=True)
class StrFactory(FactoryCore):
    class _Constructor(ConstructorCore[str]):
        def __call__(self, data: Value) -> str:
            if not isinstance(data.raw, str):
                raise data.make_error("expected string")
            return data.raw

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not str:
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())


@final
@dataclass(frozen=True)
class PrimitiveListFactory(FactoryCore):
    class _Constructor(ConstructorCore[List[Primitive]]):
        def __call__(self, data: Value) -> List[Primitive]:
            return data.as_array().raw

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not list:
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())


@final
@dataclass(frozen=True)
class PrimitiveDictFactory(FactoryCore):
    class _Constructor(ConstructorCore[Dict[str, Primitive]]):
        def __call__(self, data: Value) -> Dict[str, Primitive]:
            return data.as_object().raw

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if target is not dict:
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())


@final
@dataclass(frozen=True)
class AnyFactory(FactoryCore):
    class _Constructor(ConstructorCore[Any]):
        def __call__(self, data: Value) -> Any:
            return data.raw

    def create_constructor(
        self, _: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if not (target is Any or target is object):
            raise UnsupportedTargetTypeError
        return cast(ConstructorCore[_T_co], type(self)._Constructor())



"""Common utility functions for dealing with generic types."""

import contextlib
import inspect
import sys
import typing
from collections import abc
from dataclasses import dataclass
from functools import partial
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

from typing_extensions import Literal, TypedDict

from terramare.errors import TerramareError

from .types import LiteralMeta, Target, TypedDictMetas


def get_base_of_generic_type(type_: Target[object]) -> Optional[Target[object]]:
    """Return the generic type of which a type_ is an instance."""
    type_ = _normalize_type(type_)
    for attribute in ("__origin__", "__class__"):
        if hasattr(type_, attribute):
            return _coalesce_type(getattr(type_, attribute))
    return None  # pragma: no cover


def get_type_parameters(from_type: Target[Any]) -> List[type]:
    """Extract the type parameters from a parametrized type."""

    def extract_type_parameters(type_: Target[Any]) -> List[type]:
        type_ = _normalize_type(type_)
        with contextlib.suppress(AttributeError):
            if type_.__args__ is not None:  # type: ignore[union-attr]
                return list(type_.__args__)  # type: ignore[union-attr]
        with contextlib.suppress(AttributeError):
            if type_.__values__ is not None:  # type: ignore[union-attr]
                return list(type_.__values__)  # type: ignore[union-attr]
        return []

    from_type = _normalize_type(from_type)
    type_params = extract_type_parameters(_normalize_type(from_type))
    if getattr(from_type, "__class__", None) is LiteralMeta:
        # Special-case Literal types, as Literal[()] != Literal[]
        return type_params
    return type_params if type_params != [()] else []


@dataclass(frozen=True)
class Parameter:
    """
    A function parameter.

    A view on 'inspect.Parameter' with convenience functions for our use case.
    """

    parameter: inspect.Parameter

    @property
    def name(self) -> str:
        return self.parameter.name

    @property
    def annotation(self) -> Any:
        return self.parameter.annotation

    def is_positional(self) -> bool:
        return self.parameter.kind in {
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.VAR_POSITIONAL,
        }

    def is_positional_only(self) -> bool:
        return self.parameter.kind in {
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.VAR_POSITIONAL,
        }

    def is_optional(self) -> bool:
        if self.parameter.default == inspect.Parameter.empty:
            return False
        return True

    def is_required(self) -> bool:
        return not (self.is_optional() or self.is_variadic())

    def is_variadic(self) -> bool:
        return self.parameter.kind in {
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.VAR_KEYWORD,
        }


@dataclass(frozen=True)
class ParameterError(TerramareError):
    """Raised when failing to determine the parameters of a callable."""

    message: str


def get_parameters(type_: Target[object]) -> List[Parameter]:
    """Retrieve the parameters of the given callable."""
    try:
        return [Parameter(p) for p in inspect.signature(type_).parameters.values()]
    except ValueError as e:
        raise ParameterError(str(e)) from e


def get_type_hints(
    type_: Target[object], params: Sequence[Parameter]
) -> Dict[str, Target[object]]:
    """Retrieve the type hints of the parameters of the given callable."""
    if hasattr(type_, "__globals__"):
        globals_ = getattr(type_, "__globals__")
    elif hasattr(type_, "__module__"):
        globals_ = vars(sys.modules[getattr(type_, "__module__")])
    else:  # pragma: no cover TODO
        globals_ = {}

    # For some reason `get_type_hints` can't handle the type returned by
    # `functools.partial` - but it _can_ handle its `__call__` method.
    if isinstance(type_, partial):
        type_ = type_.__call__

    try:
        type_hints = typing.get_type_hints(type_, globals_)
    except TypeError as e:  # pragma: no cover
        raise ParameterError(str(e)) from None

    def get_type_hint(p: Parameter) -> Any:
        if p.name in type_hints:
            return type_hints[p.name]
        if p.annotation == inspect.Parameter.empty:
            return Any
        if not isinstance(p.annotation, str):
            return p.annotation
        return eval(  # pragma: no cover, pylint: disable=eval-used
            p.annotation, globals_
        )

    return {p.name: get_type_hint(p) for p in params}


def _normalize_type(type_: Target[object]) -> Target[object]:
    type_ = _coalesce_type(type_)
    try:
        return cast(
            Target[object],
            {
                AbstractSet: AbstractSet[Any],
                Callable: Callable[..., Any],
                Dict: Dict[str, Any],
                FrozenSet: FrozenSet[Any],
                Iterable: Iterable[Any],
                Iterator: Iterator[Any],
                List: List[Any],
                Mapping: Mapping[str, Any],
                MutableMapping: Mapping[str, Any],
                MutableSequence: MutableSequence[Any],
                Sequence: Sequence[Any],
                Set: Set[Any],
                Tuple: Tuple[Any, ...],
            }[type_],
        )
    except (KeyError, TypeError):
        return type_


def _coalesce_type(type_: Target[object]) -> Target[object]:
    try:
        return cast(
            Target[object],
            {
                abc.Callable: Callable,
                abc.Iterable: Iterable,
                abc.Iterator: Iterator,
                abc.Mapping: Mapping,
                abc.MutableMapping: MutableMapping,
                abc.MutableSequence: MutableSequence,
                abc.Sequence: Sequence,
                abc.Set: Set,
                dict: Dict,
                frozenset: FrozenSet,
                list: List,
                LiteralMeta: Literal,
                set: Set,
                tuple: Tuple,
                **{meta: TypedDict for meta in TypedDictMetas},
            }[type_],
        )
    except (KeyError, TypeError):
        return type_



"""Deserializer for a newtype alias."""

from typing import Any, NewType, TypeVar

from .core import ConstructorCore, FactoryCore, InternalFactory
from .errors import UnsupportedTargetTypeError
from .types import Target

_T_co = TypeVar("_T_co", covariant=True)


class NewTypeFactory(FactoryCore):
    def create_constructor(
        self, factory: "InternalFactory", target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if not _is_newtype(target):
            raise UnsupportedTargetTypeError
        return factory.create_constructor(getattr(target, "__supertype__")).unwrap()


def _is_newtype(target: Target[Any]) -> bool:
    if isinstance(NewType, type):  # pragma: no cover
        # In Python 3.10+ NewType is a class, so we can use isinstance to
        # determine whether a type is a NewType.
        return isinstance(  # pylint: disable=isinstance-second-argument-not-valid-type
            target, NewType
        )
    else:
        # Before Python 3.10 NewType is a function. We have to use the qualified
        # name of the type to figure out whether it is a NewType.
        return getattr(target, "__qualname__", None) == "NewType.<locals>.new_type"



"""Primitive wrapper types which maintain context during construction."""

import collections.abc
import copy
import dataclasses
from dataclasses import dataclass, field
from typing import (
    TYPE_CHECKING,
    ClassVar,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from typing_extensions import final

from .errors import ConstructorError
from .pretty_printer import print_primitive, print_table
from .types import ArrayPrimitive, ObjectPrimitive, Primitive


@dataclass(frozen=True)
class ConstructionConfig:
    coerce_strings: bool = False


@dataclass(frozen=True)
class Context(Mapping[str, object]):
    """
    Arbitrary user-provided construction context.

    The context is provided to :any:`terramare.structure` and is made available
    to classes during construction via parameters of this type.
    """

    EMPTY: ClassVar["Context"]

    _inner: Mapping[str, object] = field(default_factory=dict)

    def __getitem__(self, __key: str) -> object:
        return self._inner.__getitem__(__key)

    def __iter__(self) -> Iterator[str]:
        return self._inner.__iter__()

    def __len__(self) -> int:
        return self._inner.__len__()


Context.EMPTY = Context()

_Self = TypeVar("_Self", bound="Value")

_Index = Union[int, str]


@dataclass(frozen=True)
class Value:
    raw: Primitive
    config: ConstructionConfig
    context: Context
    _parent: Optional[Tuple[_Index, "Value"]]
    _type_stack: Sequence[str] = field(default_factory=list)

    @classmethod
    def new(
        cls: Type[_Self],
        data: Primitive,
        *,
        config: Optional[ConstructionConfig] = None,
        context: Context = Context.EMPTY,
    ) -> _Self:
        return cls(data, config or ConstructionConfig(), context, None)

    def is_array(self) -> bool:
        return isinstance(self.raw, list)

    def as_array(self) -> "Array":
        if not self.is_array():
            raise self.make_error("expected array")
        return Array(
            cast(ArrayPrimitive, self.raw),
            self.config,
            self.context,
            self._parent,
            self._type_stack,
        )

    def is_object(self) -> bool:
        return isinstance(self.raw, dict)

    def as_object(self) -> "Object":
        if not self.is_object():
            raise self.make_error("expected object")
        return Object(
            cast(ObjectPrimitive, self.raw),
            self.config,
            self.context,
            self._parent,
            self._type_stack,
        )

    def push_type(self: _Self, context: str) -> _Self:
        return dataclasses.replace(self, _type_stack=[*self._type_stack, context])

    def clone(self: _Self) -> _Self:
        return copy.deepcopy(self)

    def make_error(self, msg: str) -> ConstructorError:
        # We are accessing private members of other instances of Value.
        # These members are not for public use but are expected to be used in
        # this way.
        # pylint: disable=protected-access
        return ConstructorError(
            f"{self._get_path()}: {msg}",
            print_table(
                [("path", "expected structure", "actual data")]
                + [
                    (v._get_path(), "/".join(v._type_stack), print_primitive(v.raw))
                    for v in self._iter_parents()
                ]
            ),
            self._get_depth(),
        )

    def _get_path(self) -> str:
        # As above, we are accessing private members of other instances of
        # ValueView.
        # pylint: disable=protected-access
        path = [v._get_index() for v in self._iter_parents()]
        return "." + ".".join(
            _quote_context(index) for index in path if index is not None
        )

    def _get_index(self) -> Optional[_Index]:
        if self._parent:
            return self._parent[0]
        return None

    def _get_depth(self) -> Tuple[int, int]:
        return (len(list(self._iter_parents())), len(self._type_stack))

    def _iter_parents(self) -> Iterator["Value"]:
        # As above, we are accessing private members of other instances of
        # ValueView.
        # pylint: disable=protected-access
        def inner() -> Iterator["Value"]:
            current: Value = self
            while current._parent:
                yield current
                _, current = current._parent
            yield current

        return reversed(list(inner()))


# See https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime.  # noqa
if TYPE_CHECKING:  # pragma: no cover
    _BaseArray = collections.abc.Sequence[Value]
else:
    _BaseArray = collections.abc.Sequence


@final
@dataclass(frozen=True)
class Array(Value, _BaseArray):
    raw: ArrayPrimitive

    @overload
    def __getitem__(self, index: int) -> Value:
        ...  # pragma: no cover

    @overload
    def __getitem__(self, index: slice) -> List[Value]:
        ...  # pragma: no cover

    def __getitem__(self, index: Union[int, slice]) -> Union[Value, List[Value]]:
        def wrap(i: int) -> Value:
            return Value(self.raw[i], self.config, self.context, (i, self))

        if isinstance(index, int):
            return wrap(index)
        return [wrap(i) for i in range(0, len(self))[index]]

    def __delitem__(self, index: int) -> None:
        del self.raw[index]

    def __len__(self) -> int:
        return len(self.raw)

    def insert(self, index: int, value: Union[Primitive, Value]) -> None:
        if isinstance(value, Value):
            value = value.raw
        self.raw.insert(index, value)


# See https://mypy.readthedocs.io/en/latest/runtime_troubles.html#using-classes-that-are-generic-in-stubs-but-not-at-runtime.  # noqa
if TYPE_CHECKING:  # pragma: no cover
    _BaseObject = collections.abc.MutableMapping[str, Value]
else:
    _BaseObject = collections.abc.MutableMapping


@final
@dataclass(frozen=True)
class Object(Value, _BaseObject):
    raw: ObjectPrimitive

    def __getitem__(self, key: str) -> Value:
        return Value(self.raw[key], self.config, self.context, (key, self))

    def __setitem__(self, key: str, value: Union[Primitive, Value]) -> None:
        if isinstance(value, Value):
            value = value.raw
        self.raw[key] = value

    def __delitem__(self, key: str) -> None:
        del self.raw[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.raw)

    def __len__(self) -> int:
        return len(self.raw)


def _quote_context(context: Union[int, str]) -> str:
    if isinstance(context, int):
        return str(context)
    return context.replace("\\", "\\\\").replace('"', '\\"')



"""Common types and re-exports."""

from typing import Any, Callable, Dict, List, Type, TypeVar, Union

from typing_extensions import Literal, Protocol

TypedDictMetas = set()

# pylint: disable=unused-import
try:  # pragma: no cover
    from typing import _TypedDictMeta  # type: ignore[attr-defined]

    TypedDictMetas.add(_TypedDictMeta)

except ImportError:  # pragma: no cover
    pass

# pylint: disable=unused-import
try:  # pragma: no cover
    from typing_extensions import _TypedDictMeta  # type: ignore[attr-defined]

    TypedDictMetas.add(_TypedDictMeta)
except ImportError:  # pragma: no cover
    pass

try:  # pragma: no cover
    from typing_extensions import _Literal  # type: ignore[attr-defined]

    LiteralMeta = _Literal
except ImportError:  # pragma: no cover
    LiteralMeta = Literal


_T_in = TypeVar("_T_in")
_T_co = TypeVar("_T_co", covariant=True)

_PrimitiveRecursor = Union[
    None,
    bool,
    int,
    float,
    str,
    List[_T_co],
    Dict[str, _T_co],
]
_Primitive0 = _PrimitiveRecursor[Any]
_Primitive1 = _PrimitiveRecursor[_Primitive0]
_Primitive2 = _PrimitiveRecursor[_Primitive1]

Primitive = _PrimitiveRecursor[_Primitive2]
ArrayPrimitive = List[_Primitive2]
ObjectPrimitive = Dict[str, _Primitive2]


Target = Union[Type[_T_co], Callable[..., _T_co]]


class Lazy(Protocol[_T_co]):
    def __call__(self) -> _T_co:
        ...  # pragma: no cover


class Decorator(Protocol):
    def __call__(self, __target: Target[_T_in]) -> Target[_T_in]:
        ...  # pragma: no cover



"""Terramare-specific metadata for attrs classes."""

import abc
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
)

from typing_extensions import final

from .core import ConstructorCore, FactoryCore, InternalFactory
from .errors import TerramareError, UnsupportedTargetTypeError
from .safe_mapping import SafeMapping
from .types import Lazy, Target

_METADATA_FIELD = "__terramare_terramare__"

_T_co = TypeVar("_T_co", covariant=True)
_T_in = TypeVar("_T_in")


@dataclass(frozen=True)
class MetadataError(TerramareError):
    message: str


_Metadata = TypeVar("_Metadata")
_Target = TypeVar("_Target", bound=Target[Any])


def write(target: _Target, key: str, value: Any) -> _Target:
    metadata = _read_from_target(target)
    if not metadata:
        metadata = {}
        setattr(target, _METADATA_FIELD, (metadata, target))
    metadata[key] = value
    return target


TargetMetadata = Sequence[Union["Metadata[Any]", Tuple[str, Any]]]


@dataclass(frozen=True)
class MetadataCollection:
    _metadata: SafeMapping[Target[Any], Mapping[str, Any]] = field(
        default_factory=SafeMapping
    )

    @classmethod
    def new(
        cls, metadata: Mapping[Target[Any], TargetMetadata]
    ) -> "MetadataCollection":
        def normalize(target_metadata: TargetMetadata) -> Iterator[Tuple[str, Any]]:
            for element in target_metadata:
                if isinstance(element, Metadata):
                    yield (element.KEY, element.value)
                else:
                    yield element

        return cls(
            SafeMapping(
                [
                    (target, dict(normalize(target_metadata)))
                    for target, target_metadata in metadata.items()
                ]
            )
        )

    def read(self, target: Target[Any], key: str, default: _Metadata) -> _Metadata:
        if target in self._metadata:
            target_metadata = self._metadata[target]
            if key in target_metadata:
                return cast(_Metadata, target_metadata[key])
        target_metadata = _read_from_target(target)
        if key in target_metadata:
            return cast(_Metadata, target_metadata[key])
        return default


class Metadata(Generic[_Metadata], abc.ABC):
    KEY: ClassVar[str]
    DEFAULT: ClassVar[_Metadata]

    @classmethod
    def read(cls, metadata: MetadataCollection, target: _Target) -> _Metadata:
        return metadata.read(target, cls.KEY, cls.DEFAULT)

    @property
    @abc.abstractmethod
    def value(self) -> _Metadata:
        ...  # pragma: no cover

    def __call__(self, target: _Target) -> _Target:
        return write(target, self.KEY, self.value)


@dataclass(frozen=True)
class with_(Metadata[Optional[Lazy[Callable[..., Any]]]]):
    """Construct the decorated class using the given function."""

    KEY: ClassVar[str] = f"{__name__}.with_"
    DEFAULT: ClassVar[Optional[Lazy[Callable[..., Any]]]] = None

    _value: Lazy[Callable[..., Any]]

    @property
    def value(self) -> Lazy[Callable[..., Any]]:
        return self._value


@final
@dataclass(frozen=True)
class MetadataFactory(FactoryCore):
    _metadata: MetadataCollection

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        _with = with_.read(self._metadata, target)
        if _with:
            return factory.create_constructor(
                _with(), context="metadata 'with'"
            ).unwrap()
        raise UnsupportedTargetTypeError


def _read_from_target(target: Target[_T_in]) -> Dict[str, Any]:
    if not hasattr(target, _METADATA_FIELD):
        return {}
    attr = getattr(target, _METADATA_FIELD)
    if not (isinstance(attr, tuple) and isinstance(attr[0], dict)):
        raise MetadataError("unexpected type metadata")
    meta, t = attr
    # Ensure that the metadata was applied to `t` itself, rather than a base
    # class of `t`.
    # This is important because the base class may define metadata that is
    # not appropriate for derived classes, such as tagged polymorphism.
    if target != t:
        return {}
    return meta



"""Default metadata for selected standard library types."""


import datetime
import ipaddress
import pathlib
import re
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Mapping, Pattern, TypeVar

import dateutil.parser
from typing_extensions import Final

from .classes import VALUE, auto
from .errors import ValidationException
from .metadata import TargetMetadata, with_
from .types import Target


@dataclass(frozen=True)
class StdlibValidationException(ValidationException):
    """Exception type raised on failure by standard library constructors."""

    msg: str


@auto(from_=VALUE)
def _make_datetime(s: str) -> datetime.datetime:
    try:
        return dateutil.parser.isoparse(s)
    except ValueError as e:
        raise StdlibValidationException("invalid ISO-8601 datetime") from e


@auto(from_=VALUE)
def _make_regex(s: str) -> Pattern[Any]:
    try:
        return re.compile(s)
    except re.error as e:
        raise StdlibValidationException(f"invalid regex: {e}") from e


@auto(from_=VALUE)
def _make_ipv4_address(s: str) -> ipaddress.IPv4Address:
    try:
        return ipaddress.IPv4Address(s)
    except ValueError as e:
        raise StdlibValidationException(f"invalid IPv4 address: {e}") from e


@auto(from_=VALUE)
def _make_ipv6_address(s: str) -> ipaddress.IPv6Address:
    try:
        return ipaddress.IPv6Address(s)
    except ValueError as e:
        raise StdlibValidationException(f"invalid IPv6 address: {e}") from e


T = TypeVar("T")


def _make_str_fn(path_t: Callable[[str], T]) -> Callable[[str], T]:
    @auto(from_=VALUE)
    def _make_path(s: str) -> T:
        return path_t(s)

    return _make_path


DEFAULT_METADATA: Final[Mapping[Target[Any], TargetMetadata]] = {
    datetime.datetime: [with_(lambda: _make_datetime)],
    Pattern: [with_(lambda: _make_regex)],
    Pattern[str]: [with_(lambda: _make_regex)],
    ipaddress.IPv4Address: [with_(lambda: _make_ipv4_address)],
    ipaddress.IPv6Address: [with_(lambda: _make_ipv6_address)],
    **{
        t: [with_(partial(_make_str_fn, t))]
        for t in [
            pathlib.Path,
            pathlib.PosixPath,
            pathlib.WindowsPath,
            pathlib.PurePath,
            pathlib.PurePosixPath,
            pathlib.PureWindowsPath,
        ]
    },
}



"""Deserializer for a generalized homogenous sequence."""

from dataclasses import dataclass
from typing import (
    AbstractSet,
    Any,
    Callable,
    FrozenSet,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableSequence,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    cast,
)

from typing_extensions import Final, Protocol, final

from . import type_utils
from .core import ConstructorCore, FactoryCore, InternalConstructor, InternalFactory
from .data import Value
from .errors import UnsupportedTargetTypeError
from .types import Target

_T_co = TypeVar("_T_co", covariant=True)

_SEQUENCE_TYPES: Final[Mapping[Target[Any], Callable[[Sequence[Any]], Any]]] = {
    AbstractSet: set,
    FrozenSet: frozenset,
    Iterable: list,
    Iterator: iter,
    List: list,
    MutableSequence: list,
    Sequence: list,
    Set: set,
}


_E_in = TypeVar("_E_in")
_E_contra = TypeVar("_E_contra", contravariant=True)


class _ConstructFn(Protocol[_E_contra, _T_co]):
    def __call__(self, __sequence: Sequence[_E_contra]) -> _T_co:
        ...  # pragma: no cover


@final
@dataclass(frozen=True)
class SequenceFactory(FactoryCore):
    # Logically _Constructor: ConstructorImpl[SequenceLike[_T_co]] for
    # SequenceLike in {Sequence, List, ...}. Representing this requires HKTs
    # which are not currently available.
    @dataclass(frozen=True)
    class _Constructor(Generic[_T_co, _E_in], ConstructorCore[_T_co]):
        _element_constructor: InternalConstructor[_E_in]
        _construct_fn: _ConstructFn[_E_in, _T_co]

        def __call__(self, data: Value) -> _T_co:
            array_data = data.as_array()
            return self._construct_fn(
                [
                    self._element_constructor(array_data[index])
                    for index in range(len(array_data))
                ]
            )

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        base = type_utils.get_base_of_generic_type(target)
        if not base or base not in _SEQUENCE_TYPES:
            raise UnsupportedTargetTypeError
        [element_type] = type_utils.get_type_parameters(target)
        return cast(
            ConstructorCore[_T_co],
            type(self)._Constructor(
                factory.create_constructor(element_type, context="element type"),
                _SEQUENCE_TYPES[base],
            ),
        )


_E = TypeVar("_E")


@final
@dataclass(frozen=True)
class TupleFactory(FactoryCore):
    @dataclass(frozen=True)
    class _TupleConstructor(ConstructorCore[Tuple[Any, ...]]):
        _element_constructors: Sequence[InternalConstructor[Any]]

        def __call__(self, data: Value) -> Tuple[Any, ...]:
            array_data = data.as_array()
            if len(array_data) < len(self._element_constructors):
                raise data.make_error(
                    f"too few elements ({len(array_data)}) - "
                    f"expected {len(self._element_constructors)}"
                )
            if len(array_data) > len(self._element_constructors):
                raise data.make_error(
                    f"too many elements ({len(array_data)}) - "
                    f"expected {len(self._element_constructors)}"
                )
            return tuple(
                self._element_constructors[index](array_data[index])
                for index in range(len(self._element_constructors))
            )

    @dataclass(frozen=True)
    class _VariadicTupleConstructor(ConstructorCore[Tuple[_E, ...]]):
        _element_constructor: InternalConstructor[_E]

        def __call__(self, data: Value) -> Tuple[_E, ...]:
            array_data = data.as_array()
            return tuple(
                self._element_constructor(array_data[index])
                for index in range(len(array_data))
            )

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if type_utils.get_base_of_generic_type(target) is not Tuple:
            raise UnsupportedTargetTypeError
        element_types = type_utils.get_type_parameters(target)
        if element_types and element_types[-1] == Ellipsis:
            constructor: ConstructorCore[Any] = type(self)._VariadicTupleConstructor(
                factory.create_constructor(element_types[0], context="tuple element 0")
            )
        else:
            constructor = type(self)._TupleConstructor(
                [
                    factory.create_constructor(
                        element_type, context=f"tuple element {index}"
                    )
                    for index, element_type in enumerate(element_types)
                ]
            )
        return constructor



"""Pretty-printer for error messages."""

import contextlib
import itertools
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    Set,
    Tuple,
    Union,
)

from typing_extensions import Literal, final

from . import type_utils
from .types import Primitive, Target


def print_primitive(
    value: Primitive,
    short: bool = False,
    max_string_length: int = 16,
    max_dict_length: int = 3,
    max_list_length: int = 3,
) -> str:
    """Pretty-print a primitive value."""
    if isinstance(value, str):
        if len(value) > max_string_length:
            value = value[: max_string_length - 3] + "..."
        return f'"{value}"'
    if not value:
        return str(value)
    if isinstance(value, dict):
        if short:
            return "{...}"
        else:
            elts = [
                print_primitive(k) + ": " + print_primitive(v, short=True)
                for k, v in value.items()
            ]
            if len(elts) > max_dict_length:
                elts = elts[: max_dict_length - 1] + ["..."]
            return f"{{{', '.join(elts)}}}"
    if isinstance(value, list):
        if short:
            return "[...]" if value else "[]"
        else:
            elts = [print_primitive(v, short=True) for v in value]
            if len(elts) > max_list_length:
                elts = elts[: max_list_length - 1] + ["..."]
            return f"[{', '.join(elts)}]"
    return str(value)


@dataclass(frozen=True)
class LoggablePrimitive:
    """Lazy formatting for Primitive."""

    _primitive: Primitive

    def __str__(self) -> str:  # pragma: no cover
        return print_primitive(self._primitive)


def print_type_name(type_: Any) -> str:
    """Return the best available user-facing name for a type."""
    overrides: Dict[Target[Any], str] = {
        dict: "dict",
        frozenset: "frozenset",
        set: "set",
        list: "list",
        tuple: "tuple",
    }
    try:
        hash(type_)
    except TypeError:  # pragma: no cover
        pass
    else:
        if type_ in overrides:
            return overrides[type_]
    base = type_utils.get_base_of_generic_type(type_)
    if base in {
        Callable,
        Dict,
        FrozenSet,
        Iterable,
        Iterator,
        Literal,
        List,
        Mapping,
        MutableMapping,
        MutableSequence,
        Sequence,
        Set,
        Tuple,
        Union,
    }:
        parameters = type_utils.get_type_parameters(type_)
        if (
            base is Union
            and len(parameters) == 2
            and parameters[-1] is type(None)  # noqa: E721
        ):
            return f"Optional[{print_type_name(parameters[0])}]"
        return (
            _get_type_name(base)
            + "["
            + ", ".join([print_type_name(t) for t in parameters])
            + "]"
        )
    return _get_type_name(type_)


@dataclass(frozen=True)
class LoggableTarget:
    """Lazy formatting for Target."""

    _target: Target[Any]

    def __str__(self) -> str:  # pragma: no cover
        return f"'{print_type_name(self._target)}'"


@final
@dataclass(frozen=True)
class LoggableArguments:
    _positional: Sequence[object]
    _keyword: Mapping[str, object]

    def __str__(self) -> str:  # pragma: no cover
        return ", ".join(
            itertools.chain(
                (str(arg) for arg in self._positional),
                (f"{key}={arg}" for key, arg in self._keyword.items()),
            )
        )


def print_table(rows: Sequence[Tuple[str, ...]]) -> str:
    if not rows:
        return ""
    col_nos = list(range(0, max(len(row) for row in rows)))
    col_widths = [
        max(len(row[col_no]) if col_no < len(row) else 0 for row in rows)
        for col_no in col_nos
    ]
    col_nos = [col_no for col_no in col_nos if col_widths[col_no]]
    lines = []
    for row in rows:
        cols = []
        for col_no in col_nos:
            col = ""
            if col_no < len(row):
                col = row[col_no]
            cols.append(f"{col:{col_widths[col_no]}}")
        lines.append("  ".join(cols).strip())
    return "\n".join(lines)


def _get_type_name(type_: Any) -> str:
    overrides = {Any: "Any", Ellipsis: "...", Literal: "Literal", Union: "Union"}
    if type_ in overrides:
        return overrides[type_]
    with contextlib.suppress(AttributeError):  # pragma: no cover TODO
        # pylint: disable=protected-access
        if isinstance(type_._name, str):  # pragma: no branch
            return type_._name
    with contextlib.suppress(AttributeError):
        if isinstance(type_.__name__, str):  # pragma: no branch
            return type_.__name__
    with contextlib.suppress(AttributeError):
        if isinstance(type_.__qualname__, str):  # pragma: no cover TODO
            return type_.__qualname__
    if hasattr(type_, "__origin__"):  # pragma: no cover TODO
        return _get_type_name(type_.__origin__)
    return str(type_)



"""MutableMapping that can handle unhashable types."""
import collections.abc
from dataclasses import dataclass, field
from typing import (
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from typing_extensions import final

_K_in = TypeVar("_K_in")
_V_in = TypeVar("_V_in")

_V_co = TypeVar("_V_co", covariant=True)


# collections.abc.Mapping (and similar collections ABCs) are not subscriptable -
# providing type arguments causes an error at runtime!
@dataclass(init=False)
class EqMapping(Generic[_K_in, _V_co], collections.abc.Mapping):  # type: ignore[type-arg]
    """Mapping using __eq__ for comparison instead of __hash__."""

    _inner: Sequence[Tuple[_K_in, _V_co]] = field(default_factory=list)

    def __init__(
        self,
        elements: Iterable[Tuple[_K_in, _V_co]] = (),
    ) -> None:
        inner: List[Tuple[_K_in, _V_co]] = []
        for k, v in elements:
            if not any(k_ == k for k_, _ in inner):
                inner.append((k, v))
        self._inner = inner

    def __getitem__(self, k: _K_in) -> _V_co:
        index = self._get_index_of(k)
        if index is None:
            raise KeyError(k)
        return self._inner[index][1]

    def __iter__(self) -> Iterator[_K_in]:
        for k, _ in self._inner:
            yield k

    def __len__(self) -> int:
        return len(self._inner)

    def _get_index_of(self, k: _K_in) -> Optional[int]:
        for i, (k_, _) in enumerate(self._inner):
            if k_ == k:
                return i
        return None


@final
@dataclass(init=False)
class EqMutableMapping(
    EqMapping[_K_in, _V_in], collections.abc.MutableMapping  # type: ignore[type-arg]
):
    """MutableMapping using __eq__ for comparison instead of __hash__."""

    _inner: List[Tuple[_K_in, _V_in]] = field(default_factory=list)

    def __setitem__(self, k: _K_in, v: _V_in) -> None:
        index = self._get_index_of(k)
        if index is None:
            index = len(self._inner)
        else:
            del self._inner[index]
        self._inner.insert(index, (k, v))

    def __delitem__(self, k: _K_in) -> None:
        index = self._get_index_of(k)
        if index is None:
            raise KeyError(k)
        del self._inner[index]


@dataclass(init=False)
class SafeMapping(Generic[_K_in, _V_co], collections.abc.Mapping):  # type: ignore[type-arg]
    """Mapping that can handle unhashable types."""

    _hashable: Mapping[_K_in, _V_co]
    _unhashable: EqMapping[_K_in, _V_co]

    def __init__(
        self,
        elements: Iterable[Tuple[_K_in, _V_co]] = (),
    ) -> None:
        hashable = {}
        unhashable = []
        for k, v in elements:
            try:
                hash(k)
            except TypeError:
                unhashable.append((k, v))
            else:
                hashable[k] = v
        self._hashable = hashable
        self._unhashable = EqMutableMapping(unhashable)

    def __getitem__(self, k: _K_in) -> _V_co:
        try:
            hash(k)
            mapping: Mapping[_K_in, _V_co] = self._hashable
        except TypeError:
            mapping = self._unhashable
        return mapping[k]

    def __iter__(self) -> Iterator[_K_in]:
        yield from self._hashable
        yield from self._unhashable

    def __len__(self) -> int:
        return len(self._hashable) + len(self._unhashable)


@final
@dataclass(init=False)
class SafeMutableMapping(
    SafeMapping[_K_in, _V_in], collections.abc.MutableMapping  # type: ignore[type-arg]
):
    """MutableMapping that can handle unhashable types."""

    _hashable: Dict[_K_in, _V_in]
    _unhashable: EqMutableMapping[_K_in, _V_in]

    def __setitem__(self, k: _K_in, v: _V_in) -> None:
        self._get_mapping(k)[k] = v

    def __delitem__(self, k: _K_in) -> None:
        del self._get_mapping(k)[k]

    def _get_mapping(self, k: _K_in) -> MutableMapping[_K_in, _V_in]:
        try:
            hash(k)
        except TypeError:
            return self._unhashable
        return self._hashable



"""Iterator utilities."""

from typing import Callable, Iterator, List, Optional, TypeVar

_T = TypeVar("_T")


def pop_all(list_: List[_T], predicate: Callable[[_T], bool]) -> List[_T]:
    """
    Pop all elements from the list matching the given predicate.

    Order is preserved and the predicate will be called exactly once per
    element.
    """
    matches: List[_T] = []
    for index in reversed(
        [index for index, element in enumerate(list_) if predicate(element)]
    ):
        matches.insert(0, list_.pop(index))
    return matches


def pop_first(list_: List[_T], predicate: Callable[[_T], bool]) -> Optional[_T]:
    """
    Pop the first element from the list matching the given predicate.

    Order is preserved and the predicate will be called at most once per
    element.

    Returns None if no matching element is found.
    """
    for index, element in enumerate(list_):
        if predicate(element):
            return list_.pop(index)
    return None


def is_empty(iterator: Iterator[object]) -> bool:
    try:
        next(iterator)
    except StopIteration:
        return True
    return False



"""Deserializer factory for a generic class."""

import dataclasses
import enum
import logging
from dataclasses import InitVar, dataclass
from typing import (
    AbstractSet,
    Any,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Final, final

from . import iterator_utils, type_utils
from .core import ConstructorCore, FactoryCore, InternalConstructor, InternalFactory
from .data import Context, Value
from .errors import InternalError, UnsupportedTargetTypeError, ValidationException
from .metadata import Metadata, MetadataCollection
from .pretty_printer import LoggableArguments, LoggableTarget
from .types import Target

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)
_T_in = TypeVar("_T_in")
_Target = TypeVar("_Target", bound=Target[Any])


@dataclass(frozen=True)
class _disallow_unknown_fields(Metadata[bool]):

    KEY: ClassVar[str] = f"{__name__}.allow_unknown_fields"
    DEFAULT: ClassVar[bool] = False

    @property
    def value(self) -> bool:
        return True


#: Raise :exc:`terramare.errors.ConstructorError` if there are fields present in
#: the primitive that are not defined by the decorated class.
disallow_unknown_fields = _disallow_unknown_fields()


@dataclass(init=False)
class handle_exception_types(Metadata[Tuple[Type[Exception], ...]]):
    """
    Handle the listed exception types when constructing an instance of the decorated class.

    An exception of one of these types will be caught and re-raised as a
    :exc:`terramare.errors.ValidationError` including the full construction
    context.

    >>> import attr
    >>> import terramare
    >>>
    >>> @terramare.handle_exception_types(ValueError)
    ... @attr.s(auto_attribs=True)
    ... class User:
    ...     id: int = attr.ib()
    ...     name: str
    ...
    ...     @id.validator
    ...     def positive(self, _, value):
    ...         if not value > 0:
    ...             raise ValueError("id must be positive!")
    >>>
    >>> terramare.structure({"id": 0, "name": "Alice"}, into=User)
    Traceback (most recent call last):
    ...
    terramare.errors.ConstructorError: .: id must be positive!
    ...
    """

    KEY: ClassVar[str] = f"{__name__}.handle_exception_types"
    DEFAULT: ClassVar[Tuple[Type[Exception], ...]] = ()

    _value: Tuple[Type[Exception], ...]

    @property
    def value(self) -> Tuple[Type[Exception], ...]:
        return self._value

    def __init__(self, *exceptions: Type[Exception]) -> None:
        self._value = exceptions


@enum.unique
class _FromType(enum.Enum):
    OBJECT = "object"
    ARRAY = "array"
    LEAF = "leaf"


FromType = AbstractSet[_FromType]

OBJECT: Final = frozenset({_FromType.OBJECT})
ARRAY: Final = frozenset({_FromType.ARRAY})
VALUE: Final = frozenset({_FromType.LEAF})

_AUTO_KEY: Final[str] = f"{__name__}.auto"


@dataclass(frozen=True)
class _auto(Metadata[Optional[FromType]]):
    KEY: ClassVar[str] = _AUTO_KEY
    DEFAULT: ClassVar[Optional[FromType]] = None

    @dataclass(frozen=True)
    class inner(Metadata[Optional[FromType]]):
        KEY: ClassVar[str] = _AUTO_KEY
        DEFAULT: ClassVar[Optional[FromType]] = None

        _value: FromType

        @property
        def value(self) -> FromType:
            return self._value

    @property
    def value(self) -> FromType:
        return OBJECT

    @overload
    def __call__(self, __enable: bool) -> "_auto.inner":
        ...  # pragma: no cover

    @overload
    def __call__(self, *, from_: FromType) -> "_auto.inner":
        ...  # pragma: no cover

    @overload
    def __call__(self, target: _Target) -> _Target:
        ...  # pragma: no cover

    def __call__(
        self,
        target: Optional[Union[bool, Target[_T_in]]] = None,
        from_: Optional[FromType] = None,
    ) -> Union["_auto.inner", Target[_T_in]]:
        if isinstance(target, bool):
            # First overload
            if target:
                return _auto.inner(OBJECT)
            return _auto.inner(frozenset())
        if from_ is not None:
            # Second overload
            return _auto.inner(from_)
        # Third overload
        assert target is not None
        return _auto.inner(self.value)(target)


#: Automatically create a constructor for the decorated class.
auto = _auto()


@final
@dataclass(frozen=True)
class ClassFactory(FactoryCore):
    _metadata: MetadataCollection

    @dataclass(frozen=True)
    class _Constructor(ConstructorCore[_T_co]):
        _from_object: Optional[ConstructorCore[_T_co]]
        _from_array: Optional[ConstructorCore[_T_co]]
        _from_value: Optional[ConstructorCore[_T_co]]
        _handle_exception_types: Tuple[Type[Exception], ...]

        def __call__(self, data: "Value") -> _T_co:
            def get_constructor() -> Optional[ConstructorCore[_T_co]]:
                if data.is_object() and self._from_object:
                    return self._from_object
                if data.is_array() and self._from_array:
                    return self._from_array
                return self._from_value

            constructor = get_constructor()
            if not constructor:
                constructors = []
                if self._from_object:
                    constructors.append("object")
                if self._from_array:
                    constructors.append("array")
                raise data.make_error(f"expected {' or '.join(constructors)}")
            try:
                return constructor(data)
            except self._handle_exception_types as e:  # pylint: disable=catching-non-exception
                raise data.make_error(  # pylint: disable=bad-exception-context
                    str(e)
                ) from e

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if not hasattr(target, "__call__"):
            raise UnsupportedTargetTypeError

        try:
            _add_type_hints(target, type_utils.get_parameters(target))
        except type_utils.ParameterError as e:
            raise factory.make_error(str(e))

        fromtypes = _get_fromtypes(self._metadata, target)
        if fromtypes is None:
            raise factory.make_error("class construction not enabled for target")
        if not fromtypes:
            raise factory.make_error("class construction disabled for target")

        from_object = None
        if OBJECT <= fromtypes:
            from_object = self._create_from_object_constructor(factory, target)

        from_array = None
        if ARRAY <= fromtypes:
            from_array = self._create_from_array_constructor(factory, target)

        from_value = None
        if VALUE <= fromtypes:
            from_value = self._create_from_value_constructor(factory, target)

        return type(self)._Constructor(
            _from_object=from_object,
            _from_array=from_array,
            _from_value=from_value,
            _handle_exception_types=(
                ValidationException,
                *handle_exception_types.read(self._metadata, target),
            ),
        )

    @dataclass(frozen=True)
    class _FromObjectConstructor(ConstructorCore[_T_co]):
        _positional: Sequence["_ContextField"]
        _keyword: Mapping[str, "_Field"]
        _var_keyword: Optional[InternalConstructor[Any]]
        # Variadic positional parameters are not used by this constructor.

        _target: Target[_T_co]
        _disallow_unknown_fields: bool

        def __call__(self, data: Value) -> _T_co:
            object_data = data.as_object()

            positional = []
            # Covered on Python 3.8+ only.
            for _ in self._positional:  # pragma: no cover
                positional.append(data.context)

            keyword = {}
            for name, field in self._keyword.items():
                if isinstance(field, _ContextField):
                    keyword[name] = data.context
                    continue
                try:
                    element = object_data.pop(name)
                except KeyError:
                    if field.required:
                        raise data.make_error(
                            f"missing required field: {name}"
                        ) from None
                    continue
                keyword[name] = field.constructor(element)

            if self._var_keyword:
                while object_data:
                    name = next(iter(object_data))
                    keyword[name] = self._var_keyword(object_data.pop(name))

            if self._disallow_unknown_fields and object_data:
                keys_str = ", ".join(f'"{key}"' for key in object_data)
                raise data.make_error(f"unknown field(s): {keys_str}")

            _log.debug(
                "Instantiating %s with arguments: %s",
                LoggableTarget(self._target),
                LoggableArguments(positional, keyword),
            )
            return self._target(*positional, **keyword)  # type: ignore[call-arg]

    def _create_from_object_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        parameters = _add_type_hints(target, type_utils.get_parameters(target))

        positional = []
        # Covered on Python 3.8+ only.
        while (
            parameters
            and parameters[0].is_positional_only()
            and not parameters[0].is_variadic()
        ):  # pragma: no cover
            parameter = parameters.pop(0)
            if parameter.is_context():
                positional.append(_ContextField())
            elif parameter.is_required():
                raise factory.make_error(
                    "cannot create from-object constructor for target with required "
                    f"positional-only parameter: {parameter.name}"
                )
            # Positional parameter is optional; omit it.

        positional_only_parameters = [
            p for p in parameters if p.is_positional_only() and not p.is_variadic()
        ]
        if positional_only_parameters:  # pragma: no cover
            raise InternalError(
                f"unexpected positional-only parameters remaining: {positional_only_parameters}"
            )

        keyword: Dict[str, _Field] = {}
        var_keyword = None
        while parameters:
            parameter = parameters.pop(0)
            if parameter.is_variadic():
                # The _FromObjectConstructor doesn't use variadic positional
                # parameters.
                if parameter.is_positional():
                    pass
                # The _FromObjectConstructor doesn't use variadic context
                # parameters.
                if not parameter.is_context():
                    var_keyword = factory.create_constructor(
                        parameter.type_hint, context=f'field "{parameter.name}"'
                    )
            elif parameter.is_context():
                keyword[parameter.name] = _ContextField()
            else:
                keyword[parameter.name] = _SingleField(
                    factory.create_constructor(
                        parameter.type_hint, context=f'field "{parameter.name}"'
                    ),
                    parameter.is_required(),
                )

        if parameters:  # pragma: no cover
            raise InternalError(f"unexpected parameters remaining: {parameters}")

        return type(self)._FromObjectConstructor(
            _positional=positional,
            _keyword=keyword,
            _var_keyword=var_keyword,
            _target=target,
            _disallow_unknown_fields=disallow_unknown_fields.read(
                self._metadata, target
            ),
        )

    @dataclass(frozen=True)
    class _FromArrayConstructor(ConstructorCore[_T_co]):
        _positional: Sequence["_Field"]
        _var_positional: Optional[InternalConstructor[Any]]
        _keyword: Mapping[str, "_ContextField"]
        # Variadic keyword parameters are not used by this constructor.

        _target: Target[_T_co]
        _disallow_extra_elements: bool

        def __call__(self, data: Value) -> _T_co:
            array_data = data.as_array()
            iter_data = iter(array_data)

            positional: List[Any] = []
            for field in self._positional:
                if isinstance(field, _ContextField):
                    positional.append(data.context)
                    continue
                try:
                    element = next(iter_data)
                except StopIteration:
                    if field.required:
                        raise data.make_error(
                            f"too few elements ({len(array_data)}) - "
                            f"expected at least {self.min_elements} "
                        ) from None
                    break
                positional.append(field.constructor(element))

            if self._var_positional:
                for element in iter_data:
                    positional.append(self._var_positional(element))

            if self._disallow_extra_elements and not iterator_utils.is_empty(iter_data):
                raise data.make_error(
                    f"too many elements ({len(array_data)}) - "
                    f"expected at most {self.max_elements}"
                )

            keyword = {field: data.context for field in self._keyword}

            _log.debug(
                "Instantiating %s with arguments: %s",
                LoggableTarget(self._target),
                LoggableArguments(positional, keyword),
            )
            return self._target(*positional, **keyword)  # type: ignore[call-arg]

        @property
        def min_elements(self) -> int:
            return sum(
                1
                for field in self._positional
                if isinstance(field, _SingleField) and field.required
            )

        @property
        def max_elements(self) -> int:
            if self._var_positional:  # pragma: no cover
                raise InternalError(
                    "no maximum element count for constructor with variadic parameter"
                )
            return sum(
                1
                for field in self._positional
                if isinstance(field, _SingleField) and field.required
            )

    def _create_from_array_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        parameters = _add_type_hints(target, type_utils.get_parameters(target))

        positional: List[_Field] = []
        var_positional = None

        while parameters and parameters[0].is_positional():
            parameter = parameters.pop(0)
            if parameter.is_variadic():
                # The _FromArrayConstructor doesn't use variadic context
                # parameters.
                if not parameter.is_context():
                    var_positional = factory.create_constructor(
                        parameter.type_hint, context=f'field "{parameter.name}"'
                    )
            elif parameter.is_context():
                positional.append(_ContextField())
            else:
                positional.append(
                    _SingleField(
                        factory.create_constructor(
                            parameter.type_hint, context=f'field "{parameter.name}"'
                        ),
                        parameter.is_required(),
                    )
                )

        positional_parameters = [p for p in parameters if p.is_positional()]
        if positional_parameters:  # pragma: no cover
            raise InternalError(
                f"unexpected positional parameters remaining: {positional_parameters}"
            )

        keyword = {}
        while parameters:
            parameter = parameters.pop(0)
            if parameter.is_variadic():
                # The _FromArrayConstructor doesn't use variadic keyword
                # parameters.
                pass
            elif parameter.is_context():
                keyword[parameter.name] = _ContextField()
            elif parameter.is_required():
                raise factory.make_error(
                    "cannot create from-array constructor for target with required "
                    f"keyword-only parameter: {parameter.name}"
                )
            # Covered, despite what 'coverage' says.
            else:  # pragma: no cover
                # Keyword parameter is optional; omit it.
                pass

        if parameters:  # pragma: no cover
            raise InternalError(f"unexpected parameters remaining: {parameters}")

        return type(self)._FromArrayConstructor(
            _positional=positional,
            _var_positional=var_positional,
            _keyword=keyword,
            _target=target,
            _disallow_extra_elements=disallow_unknown_fields.read(
                self._metadata, target
            ),
        )

    @dataclass(frozen=True)
    class _FromValueConstructor(ConstructorCore[_T_co]):
        _positional: Sequence["_Field"]
        _keyword: Mapping[str, "_ContextField"]
        # Variadic parameters are not used by this constructor.

        _target: Target[_T_co]

        def __post_init__(self) -> None:
            values = [f for f in self._positional if isinstance(f, _SingleField)]
            if len(values) != 1:  # pragma: no cover
                raise InternalError(f"expected exactly one value field: {values}")

        def __call__(self, data: Value) -> _T_co:
            positional = []
            for field in self._positional:
                if isinstance(field, _ContextField):
                    positional.append(data.context)
                else:
                    positional.append(field.constructor(data))
            keyword = {field: data.context for field in self._keyword}
            _log.debug(
                "Instantiating %s with argument: %s",
                LoggableTarget(self._target),
                LoggableArguments(positional, keyword),
            )
            return self._target(*positional, **keyword)  # type: ignore[call-arg]

    def _create_from_value_constructor(
        self,
        factory: InternalFactory,
        target: Target[_T_co],
    ) -> ConstructorCore[_T_co]:
        parameters = _add_type_hints(target, type_utils.get_parameters(target))

        positional: List[_Field] = []

        def have_value() -> bool:
            return any(isinstance(f, _SingleField) for f in positional)

        while parameters and parameters[0].is_positional():
            parameter = parameters.pop(0)
            if parameter.is_context():
                if parameter.is_variadic():
                    # The _FromValueConstructor doesn't use variadic context
                    # parameters.
                    break
                positional.append(_ContextField())
            elif not have_value():
                positional.append(
                    _SingleField(
                        factory.create_constructor(
                            parameter.type_hint, context=f'field "{parameter.name}"'
                        ),
                        parameter.is_required(),
                    )
                )
            elif parameter.is_required():
                raise factory.make_error(
                    "cannot create from-value constructor for target "
                    "with more than one required parameter"
                )
            # Covered, despite what 'coverage' says.
            else:  # pragma: no cover
                # Tail optional parameter; omit.
                pass

        if not have_value():
            raise factory.make_error(
                "cannot create from-value constructor for "
                "target with no positional parameters"
            )

        positional_parameters = [p for p in parameters if p.is_positional()]
        if positional_parameters:  # pragma: no cover
            raise InternalError(
                f"unexpected positional parameters remaining: {positional_parameters}"
            )

        keyword = {}
        while parameters:
            parameter = parameters.pop(0)
            if parameter.is_variadic():
                # The _FromValueConstructor doesn't use variadic parameters.
                break
            if parameter.is_context():
                keyword[parameter.name] = _ContextField()
            elif parameter.is_required():
                raise factory.make_error(
                    "cannot create from-value constructor for "
                    "target with required keyword-only parameter"
                )
            # Keyword parameter is optional; omit it.

        if parameters:  # pragma: no cover
            raise InternalError(f"unexpected parameters remaining: {parameters}")

        return type(self)._FromValueConstructor(
            _positional=positional, _keyword=keyword, _target=target
        )


def _get_fromtypes(
    metadata: MetadataCollection, target: Target[Any]
) -> Optional[FromType]:
    from_types = auto.read(metadata, target)
    if from_types is not None:
        return from_types
    if dataclasses.is_dataclass(target) or hasattr(target, "__attrs_attrs__"):
        return OBJECT
    if _is_namedtuple(target):
        return OBJECT | ARRAY
    return None


def _is_namedtuple(target: Target[_T_co]) -> bool:
    if getattr(target, "__bases__", None) != (tuple,):
        return False
    fields = getattr(target, "_fields", None)
    return isinstance(fields, tuple) and all(isinstance(field, str) for field in fields)


@dataclass(frozen=True)
class _TypedParameter(type_utils.Parameter):
    type_hint: Target[object]

    def is_context(self) -> bool:
        return self.type_hint is Context


def _add_type_hints(
    type_: Target[object], params: Sequence[type_utils.Parameter]
) -> List[_TypedParameter]:
    type_hints = type_utils.get_type_hints(type_, params)
    return [
        _TypedParameter(p.parameter, _strip_initvar(type_hints[p.name])) for p in params
    ]


def _strip_initvar(target: Target[object]) -> Target[object]:
    # One or other of these branches is covered, depending on the Python
    # version.
    if isinstance(target, InitVar) and hasattr(target, "type"):  # pragma: no cover
        return target.type
    if target is InitVar:  # pragma: no cover
        raise type_utils.ParameterError(
            "'InitVar' is not supported before Python 3.8 as it does not "
            "contain the necessary type information - see "
            "https://bugs.python.org/issue33569."
        )
    return target


@dataclass(frozen=True)
class _SingleField:
    constructor: InternalConstructor[Any]
    required: bool


@dataclass(frozen=True)
class _ContextField:
    pass


_Field = Union[_SingleField, _ContextField]



"""Automatically construct complex objects from simple Python types."""

from .classes import (
    ARRAY,
    OBJECT,
    VALUE,
    auto,
    disallow_unknown_fields,
    handle_exception_types,
)
from .data import Context
from .errors import ConstructorError, FactoryError, TerramareError
from .metadata import with_
from .tagged_constructors import externally_tagged, internally_tagged
from .terramare import structure

# Expose the package version as __version__ (see PEP 396).
__version__ = "0.5.4"

__all__ = [
    "ARRAY",
    "OBJECT",
    "VALUE",
    "ConstructorError",
    "Context",
    "FactoryError",
    "TerramareError",
    "auto",
    "disallow_unknown_fields",
    "externally_tagged",
    "handle_exception_types",
    "internally_tagged",
    "structure",
    "with_",
]



"""Composite factory."""

import logging
from dataclasses import dataclass, field
from typing import Sequence, TypeVar

from typing_extensions import final

from .core import ConstructorCore, FactoryCore, InternalFactory
from .errors import UnsupportedTargetTypeError
from .pretty_printer import print_type_name
from .types import Target

_log = logging.getLogger(__name__)

_T_co = TypeVar("_T_co", covariant=True)


@final
@dataclass(frozen=True)
class CompositeFactory(FactoryCore):
    _factories: Sequence[FactoryCore] = field(default_factory=list)

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> "ConstructorCore[_T_co]":
        for f in self._factories:
            try:
                _log.debug("Attempting to create constructor using factory %s", f)
                constructor = f.create_constructor(factory, target)
                _log.debug("Successfully created constructor using factory %s", f)
                return constructor
            except UnsupportedTargetTypeError:
                _log.debug("Constructor mismatch for factory %s", f)
        raise factory.make_error(
            f"unable to construct type '{print_type_name(target)}'"
        )



"""Automatically deserialize complex objects from simple Python types."""

import logging
from typing import Any, Mapping, Optional, TypeVar, overload

from typing_extensions import Final

from .core import Factory
from .types import Primitive, Target

_log = logging.getLogger(__name__)

_DEFAULT_CONSTRUCTOR_FACTORY: Final[Factory] = Factory()

_T_co = TypeVar("_T_co", covariant=True)


@overload
def structure(
    data: Primitive,
    *,
    into: Target[_T_co],
    coerce_strings: bool = False,
    context: Optional[Mapping[str, object]] = None
) -> _T_co:
    # This overload is required to provide the correct return type for types
    # that are considered valid values of Type[T] (for some T) by mypy.
    ...  # pragma: no cover


@overload
def structure(
    data: Primitive,
    *,
    into: Any,
    coerce_strings: bool = False,
    context: Optional[Mapping[str, object]] = None
) -> Any:
    # This overload is required as 'typing' types, such as Tuple[int],
    # are not considered by mypy to be valid values of Type[T] for any T.
    ...  # pragma: no cover


def structure(
    data: Primitive,
    *,
    into: Target[Any],
    coerce_strings: bool = False,
    context: Optional[Mapping[str, object]] = None
) -> Any:
    """
    Construct a primitive into an instance of the specified target type.

    :param data: Primitive value from which to construct an instance of the
        target type.
    :param into: Target type.
    :param coerce_strings: If set, attempt to convert :python:`str` values to
        :python:`bool`, :python:`int`, or :python:`float` where a value of one
        of these types is required.

        >>> structure("1", into=int)
        Traceback (most recent call last):
            ...
        terramare.errors.ConstructorError: ...
        >>> structure("1", into=int, coerce_strings=True)
        1

        Note that setting this option will cause :python:`terramare` to reject
        non-string primitives where a :python:`bool`, :python:`int`, or :python:`float`
        is required.
        For example:

        >>> structure(1, into=int)
        1
        >>> structure(1, into=int, coerce_strings=True)
        Traceback (most recent call last):
            ...
        terramare.errors.ConstructorError: ...

        This option defaults to :python:`False`.

        :raises: :exc:`terramare.errors.FactoryError`: If a constructor for the
            target type cannot be created.
        :raises: :exc:`terramare.errors.ConstructorError`: If the primitive
            value has an incorrect type, incorrect structure, or fails value-
            level validation.
    """
    _log.debug("Using global default constructor factory")
    return _DEFAULT_CONSTRUCTOR_FACTORY.structure(
        data, into=into, coerce_strings=coerce_strings, context=context
    )



"""Common error types."""

from dataclasses import dataclass
from typing import Tuple


class TerramareError(Exception):
    """Base class for all exceptions raised by terramare."""


@dataclass(frozen=True)
class ConstructorError(TerramareError):
    """Raised when failing to construct an instance of a class from a primitive value."""

    summary: str
    detail: str = ""
    _depth: Tuple[int, int] = (0, 0)

    def __str__(self) -> str:
        return f"{self.summary}\n{self.detail}"

    def __lt__(self, other: "ConstructorError") -> bool:
        return self._depth < other._depth


class ValidationException(TerramareError):
    """Base class for field validation errors automatically caught by terramare."""


@dataclass(frozen=True)
class FactoryError(TerramareError):
    """Raised when failing to create a constructor for a type."""

    message: str


@dataclass(frozen=True)
class InternalError(TerramareError):
    """Raised on unexpected internal errors."""

    message: str

    def __str__(self) -> str:  # pragma: no cover
        return f"an internal error occured: {self.message}"


@dataclass(frozen=True)
class UnsupportedTargetTypeError(TerramareError):
    """Raised when a factory is called with a target type it does not support."""



"""Deserializers for typed dictionaries."""

from dataclasses import dataclass
from typing import Any, Dict, Mapping, MutableMapping, TypeVar, cast

from typing_extensions import TypedDict, final

from . import type_utils
from .core import ConstructorCore, FactoryCore, InternalConstructor, InternalFactory
from .data import Value
from .errors import UnsupportedTargetTypeError
from .types import Target

_T_co = TypeVar("_T_co", covariant=True)
_V_in = TypeVar("_V_in")


@final
@dataclass(frozen=True)
class MappingFactory(FactoryCore):
    # Logically _Constructor: ConstructorImpl[MappingLike[_T_co]] for
    # SequenceLike in {Mapping, Dict, ...}. Representing this requires HKTs
    # which are not currently available.
    @dataclass(frozen=True)
    class _Constructor(ConstructorCore[Dict[str, _V_in]]):
        _value_constructor: InternalConstructor[_V_in]

        def __call__(self, data: Value) -> Dict[str, _V_in]:
            object_data = data.as_object()
            return {
                key: self._value_constructor(object_data[key]) for key in object_data
            }

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if type_utils.get_base_of_generic_type(target) not in {
            Dict,
            Mapping,
            MutableMapping,
        }:
            raise UnsupportedTargetTypeError
        [key_type, value_type] = type_utils.get_type_parameters(target)
        if key_type is not str:
            raise factory.make_error("only dictionary keys of type str are supported")
        return cast(
            ConstructorCore[_T_co],
            type(self)._Constructor(
                factory.create_constructor(value_type, context="value type"),
            ),
        )


@final
@dataclass(frozen=True)
class TypedDictFactory(FactoryCore):
    @dataclass(frozen=True)
    class _Constructor(ConstructorCore[Dict[str, Any]]):
        fields: Mapping[str, InternalConstructor[Any]]
        total: bool

        def __call__(self, data: Value) -> Dict[str, Any]:
            object_data = data.as_object()
            result: Dict[str, Any] = {}

            for field, constructor in self.fields.items():
                if field in object_data:
                    result[field] = constructor(object_data[field])
                elif self.total:
                    raise data.make_error(f"missing required field: {field}")

            unused_fields = set(object_data) - set(result)
            if unused_fields:
                raise data.make_error(f"unknown field: {sorted(unused_fields)[0]}")
            return object_data.raw

    def create_constructor(
        self, factory: InternalFactory, target: Target[_T_co]
    ) -> ConstructorCore[_T_co]:
        if type_utils.get_base_of_generic_type(target) is not TypedDict:
            raise UnsupportedTargetTypeError
        return cast(
            ConstructorCore[_T_co],
            type(self)._Constructor(
                {
                    field: factory.create_constructor(
                        field_type, context=f'field "{field}"'
                    )
                    for field, field_type in target.__annotations__.items()
                },
                getattr(target, "__total__"),
            ),
        )



"""Factory preventing deserialization of Context type."""

from dataclasses import dataclass
from typing import NoReturn

from typing_extensions import final

from .core import FactoryCore, InternalFactory
from .data import Context
from .errors import UnsupportedTargetTypeError
from .types import Target


@final
@dataclass(frozen=True)
class ContextFactory(FactoryCore):
    def create_constructor(
        self, factory: InternalFactory, target: Target[object]
    ) -> NoReturn:
        if not (isinstance(target, type) and issubclass(target, Context)):
            raise UnsupportedTargetTypeError
        raise factory.make_error(
            "construction explicitly disabled for this type. "
            " You may be using the `Context` type in an incorrect location."
        )
